#include <chrono>
#include <memory>
#include <string>

#include "rclcpp/rclcpp.hpp"
#include "std_msgs/msg/string.hpp"
#include "sensor_msgs/msg/compressed_image.hpp"
#include "sensor_msgs/msg/camera_info.hpp"
#include "cv_bridge/cv_bridge.h"

#include "lifecycle_msgs/msg/transition.hpp"
#include "lifecycle_msgs/msg/state.hpp"
#include "rclcpp/publisher.hpp"
#include "rclcpp_lifecycle/lifecycle_node.hpp"
#include "rclcpp_lifecycle/lifecycle_publisher.hpp"

#include <opencv2/opencv.hpp>
#include "cameras.hpp"

using namespace std::chrono_literals;

struct Config
{
  std::string namespace_;
  std::string type;
  std::string src;
  bool enable_depth;
  int height;
  int width;
  int rate;
  std::string fourcc;
  std::string camera_info_url;
};

// const int INPUT_IMAGE_HEIGHT = 720;
// const int INPUT_IMAGE_WIDTH = 1280;

void vertical_to_horizontal(cv::Mat& frame) {
    int h = frame.rows;
    int w = frame.cols;

    double target_ratio = 1.3;
    double ratio = static_cast<double>(w) / h;

    if (ratio < target_ratio) {
        int new_h = static_cast<int>(w / target_ratio);
        int crop = static_cast<int>((h - new_h) / 2);
        frame = frame(cv::Rect(0, crop, w, new_h));
    }
}

// const rclcpp::QoS DATA_QOS = rclcpp::SensorDataQoS();
const rclcpp::QoS DATA_QOS = 5;

class CameraNode : public rclcpp_lifecycle::LifecycleNode
{
public:
  explicit CameraNode()
      : LifecycleNode("camera_node")
  {
    declare_parameter("namespace", "camera");
    declare_parameter("rate", 30);
    declare_parameter("type", "usb");
    declare_parameter("src", "0");
    declare_parameter("enable_depth", false);
    declare_parameter("height", 480);
    declare_parameter("width", 640);
    declare_parameter("fourcc", "MJPG");
    declare_parameter("camera_info_url", "");

    get_parameters();
    
    config_update_callback_handler_ = add_on_set_parameters_callback(
        std::bind(&CameraNode::config_update, this, std::placeholders::_1));

    image_pub_ = create_publisher<sensor_msgs::msg::Image>(
        config_.namespace_ + "/image_raw", DATA_QOS);
    camera_info_pub_ = create_publisher<sensor_msgs::msg::CameraInfo>(
        config_.namespace_ + "/camera_info", DATA_QOS);

    depth_pub_ = create_publisher<sensor_msgs::msg::Image>(
        config_.namespace_ + "/image_raw/depth", DATA_QOS);
    gray_pub_ = create_publisher<sensor_msgs::msg::Image>(
        config_.namespace_ + "/image_raw/gray", DATA_QOS);
    rgbd_pub_ = create_publisher<sensor_msgs::msg::Image>(
        config_.namespace_ + "/image_raw/rgbd", DATA_QOS);
    RCLCPP_INFO(get_logger(), "CameraNode has been initialized");
  }

private:
  rclcpp::Node::OnSetParametersCallbackHandle::SharedPtr config_update_callback_handler_;
  Config config_;
  std::mutex lock_;
  CameraInterface *camera_;
  cv::Mat frame_;
  cv::Mat depth_;
  cv::Mat gray_;
  cv::Mat rgbd_;
  int frame_id_;
  rclcpp_lifecycle::LifecyclePublisher<sensor_msgs::msg::Image>::SharedPtr image_pub_;
  rclcpp_lifecycle::LifecyclePublisher<sensor_msgs::msg::CameraInfo>::SharedPtr camera_info_pub_;

  rclcpp_lifecycle::LifecyclePublisher<sensor_msgs::msg::Image>::SharedPtr depth_pub_;
  rclcpp_lifecycle::LifecyclePublisher<sensor_msgs::msg::Image>::SharedPtr gray_pub_;
  rclcpp_lifecycle::LifecyclePublisher<sensor_msgs::msg::Image>::SharedPtr rgbd_pub_;
  // rclcpp_lifecycle::LifecyclePublisher<sensor_msgs::msg::CompressedImage>::SharedPtr compressed_image_pub_;

  sensor_msgs::msg::CameraInfo camera_info_;
  rclcpp::TimerBase::SharedPtr timer_;
  cv_bridge::CvImage cv_image_;

  rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn
  on_configure(const rclcpp_lifecycle::State &);
  rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn
  on_cleanup(const rclcpp_lifecycle::State &);

  rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn
  on_activate(const rclcpp_lifecycle::State &);
  rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn
  on_deactivate(const rclcpp_lifecycle::State &);
  rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn
  on_shutdown(const rclcpp_lifecycle::State &);

  rcl_interfaces::msg::SetParametersResult
  config_update(const std::vector<rclcpp::Parameter> &params);
  void clean_camera();

  void restart_camera();

  void capture_frames();

  void get_parameters();
};