#include <chrono>
#include <thread>
#include "tof/utils.hpp"
#include "v4l2wrapper/V4l2Capture.h"
#include "formats/mpeg2bgr.hpp"
#include <unistd.h>

class CameraInterface
{
public:
  CameraInterface() {}
  virtual ~CameraInterface() {}
  virtual void start() = 0;
  virtual bool read(cv::Mat &frame) = 0;
  virtual bool read(cv::Mat &frame, cv::Mat &depth, cv::Mat &gray, cv::Mat &rgbd) = 0;
  virtual void release() = 0;
  virtual bool isOpened() const = 0;
  virtual std::string getCameraType() const = 0;
};

class USBCamera : public CameraInterface
{
private:
  cv::VideoCapture camera_;
  cv::Mat frame_;
  std::string fourcc_;
  int width_;
  int height_;
  int fps_;

public:
  USBCamera(int index, std::string fourcc, int width, int height, int fps);
  void start();
  bool read(cv::Mat &frame);
  bool read(cv::Mat &frame, cv::Mat &depth, cv::Mat &gray, cv::Mat &rgbd);
  std::string getCameraType() const;
  void release();
  bool isOpened() const;
};

class V4L2Camera : public CameraInterface
{
private:
  V4l2Capture *camera_;
  cv::Mat frame_;
  std::mutex mutex_;
  std::thread *read_thread_;
  bool running_;
  char *frame_buffer_;
  MJPEG2BGR *mjpeg2bgr_;
  int width_;
  int height_;
  int fps_;

public:
  V4L2Camera(int index, int width, int height, int fps);
  void start();
  void read_loop();
  bool read(cv::Mat &frame);
  bool read(cv::Mat &frame, cv::Mat &depth, cv::Mat &gray, cv::Mat &rgbd);
  std::string getCameraType() const;
  void release();
  bool isOpened() const;
};

class VideoFile : public CameraInterface
{
private:
  cv::VideoCapture camera_;
  cv::Mat frame_;

public:
  VideoFile(std::string path);
  void start();
  bool read(cv::Mat &frame);
  bool read(cv::Mat &frame, cv::Mat &depth, cv::Mat &gray, cv::Mat &rgbd);
  std::string getCameraType() const;
  void release();
  bool isOpened() const;
};

class TOFCamera : public CameraInterface
{
private:
  HTOFD hTofD = NULL;

  TofDeviceInfo struCaps;
  TofDeviceCapability *pCaps;
  StreamCallBackParam *pStreamParam;
  bool enable_depth_;

public:
  TOFCamera(bool enable_depth);
  ~TOFCamera();
  bool isOpened() const;
  void start();
  void release();
  bool read(cv::Mat &frame);
  bool read(cv::Mat &frame, cv::Mat &depth, cv::Mat &gray, cv::Mat &rgbd);
  std::string getCameraType() const;
};


static bool isTOFCamera(cv::VideoCapture cap)
{
  int height = cap.get(cv::CAP_PROP_FRAME_HEIGHT);
  int width = cap.get(cv::CAP_PROP_FRAME_WIDTH);
  // std::cout << i << " " << height << " " << width << std::endl;
  if (height == 1460 && width == 224) // TOF Camera size
  {
    return  true;
  }
  return false;
}

static int checkCameraIndex(int index)
{
  char devName[16];
  cv::VideoCapture cap;
  cap.open(index, cv::CAP_V4L2);
  if (cap.isOpened())
  {
    bool isTOF = isTOFCamera(cap);
    cap.release();
    if (!isTOF){
      printf("isTOF: %d\n", isTOF);
      return index;
    }
  }
  cap.release();

  // glob /dev/video* to find all usb camera
  std::vector<cv::String> filenames;
  cv::glob("/dev/video*", filenames);
  for (int i = 0; i < int(filenames.size()); i++)
  {
    cap.open(i, cv::CAP_V4L2);
    if (cap.isOpened())
    {
      bool isTOF = isTOFCamera(cap);
      cap.release();
      if (!isTOF){
        return i;
      }
    }
    else
    {
      cap.release();
    }
  }
  return -1;
}
