#include <iostream>

extern "C" {
#define __STDC_CONSTANT_MACROS  // Required for libavutil
#include "libavutil/imgutils.h"
#include "libavformat/avformat.h"
#include "libavutil/error.h"
#include "libavutil/log.h"
#include "linux/videodev2.h"
#include "libswscale/swscale.h"
}
#include "av_pixel_format_helper.hpp"

class MJPEG2BGR
{
public:
  explicit MJPEG2BGR(const int width, const int height, const std::string & av_device_format_str):
    m_avcodec(avcodec_find_decoder(AVCodecID::AV_CODEC_ID_MJPEG)),
    m_avparser(av_parser_init(AVCodecID::AV_CODEC_ID_MJPEG)),
    m_avframe_device(av_frame_alloc()),
    m_avframe_bgr(av_frame_alloc()),
    m_avoptions(NULL),
    m_averror_str(reinterpret_cast<char *>(malloc(AV_ERROR_MAX_STRING_SIZE)))
  {
    if (!m_avcodec) {
      throw std::runtime_error("Could not find MJPEG decoder");
    }

    if (!m_avparser) {
      throw std::runtime_error("Could not find MJPEG parser");
    }

    m_avcodec_context = avcodec_alloc_context3(m_avcodec);

    m_avframe_device->width = width;
    m_avframe_device->height = height;
    m_avframe_device->format = AV_PIX_FMT_YUV422P;
    m_avframe_device->format = usb_cam::formats::get_av_pixel_format_from_string(av_device_format_str);

    m_avframe_bgr->width = width;
    m_avframe_bgr->height = height;
    m_avframe_bgr->format = AV_PIX_FMT_BGR24;

    m_sws_context = sws_getContext(
      width, height, (AVPixelFormat)m_avframe_device->format,
      width, height, (AVPixelFormat)m_avframe_bgr->format, SWS_FAST_BILINEAR,
      NULL, NULL, NULL);

    // Suppress warnings from ffmpeg libraries to avoid spamming the console
    av_log_set_level(AV_LOG_FATAL);
    av_log_set_flags(AV_LOG_SKIP_REPEATED);
    // av_log_set_flags(AV_LOG_PRINT_LEVEL);

    m_avcodec_context->width = width;
    m_avcodec_context->height = height;
    m_avcodec_context->pix_fmt = (AVPixelFormat)m_avframe_device->format;
    m_avcodec_context->codec_type = AVMEDIA_TYPE_VIDEO;

    m_avframe_device_size = static_cast<size_t>(
      av_image_get_buffer_size(
        (AVPixelFormat)m_avframe_device->format,
        m_avframe_device->width,
        m_avframe_device->height,
        m_align));
    m_avframe_bgr_size = static_cast<size_t>(
      av_image_get_buffer_size(
        (AVPixelFormat)m_avframe_bgr->format,
        m_avframe_bgr->width,
        m_avframe_bgr->height,
        m_align));

    // Initialize AVCodecContext
    if (avcodec_open2(m_avcodec_context, m_avcodec, &m_avoptions) < 0) {
      throw std::runtime_error("Could not open decoder");
      return;
    }

    m_result = av_frame_get_buffer(m_avframe_device, m_align);
    if (m_result != 0) {
      print_av_error_string(m_result);
    }
    m_result = av_frame_get_buffer(m_avframe_bgr, m_align);
    if (m_result != 0) {
      print_av_error_string(m_result);
    }
  }

  ~MJPEG2BGR()
  {
    if (m_averror_str) {
      free(m_averror_str);
    }
    if (m_avoptions) {
      free(m_avoptions);
    }
    if (m_avcodec_context) {
      avcodec_close(m_avcodec_context);
      avcodec_free_context(&m_avcodec_context);
    }
    if (m_avframe_device) {
      av_frame_free(&m_avframe_device);
    }
    if (m_avframe_bgr) {
      av_frame_free(&m_avframe_bgr);
    }
    if (m_avparser) {
      av_parser_close(m_avparser);
    }

    if (m_sws_context) {
      sws_freeContext(m_sws_context);
    }
  }

  void convert(const char *  src, char * dest, const int & bytes_used)
  {
    // printf("m_avframe_device_size: %d\n", m_avframe_device_size);
    // printf("bytes_used: %d\n", bytes_used);
    m_result = 0;
    // clear the picture
    memset(dest, 0, m_avframe_device_size);
    // printf("bytes_used: %d\n", bytes_used);
    auto avpacket = av_packet_alloc();
    av_new_packet(avpacket, bytes_used);
    memcpy(avpacket->data, src, bytes_used);

    // Pass src MJPEG image to decoder
    m_result = avcodec_send_packet(m_avcodec_context, avpacket);

    av_packet_free(&avpacket);

    // If result is not 0, report what went wrong
    if (m_result != 0) {
      std::cerr << "Failed to send AVPacket to decode: ";
      print_av_error_string(m_result);
      return;
    }

    m_result = avcodec_receive_frame(m_avcodec_context, m_avframe_device);

    if (m_result == AVERROR(EAGAIN) || m_result == AVERROR_EOF) {
      return;
    } else if (m_result < 0) {
      std::cerr << "Failed to recieve decoded frame from codec: ";
      print_av_error_string(m_result);
      return;
    }

    sws_scale(
      m_sws_context, m_avframe_device->data,
      m_avframe_device->linesize, 0, m_avframe_device->height,
      m_avframe_bgr->data, m_avframe_bgr->linesize);

    av_image_copy_to_buffer(
      const_cast<uint8_t *>(reinterpret_cast<const uint8_t *>(dest)),
      m_avframe_bgr_size, m_avframe_bgr->data,
      m_avframe_bgr->linesize, (AVPixelFormat)m_avframe_bgr->format,
      m_avframe_bgr->width, m_avframe_bgr->height, m_align);
  }

private:
  void print_av_error_string(int & err_code)
  {
    av_make_error_string(m_averror_str, AV_ERROR_MAX_STRING_SIZE, err_code);
    std::cerr << m_averror_str << std::endl;
  }

  AVCodec * m_avcodec;
  AVCodecContext * m_avcodec_context;
  AVCodecParserContext * m_avparser;
  AVFrame * m_avframe_device;
  AVFrame * m_avframe_bgr;
  AVDictionary * m_avoptions;
  SwsContext * m_sws_context;
  size_t m_avframe_device_size;
  size_t m_avframe_bgr_size;
  char * m_averror_str;
  int m_result = 0;
  int m_counter = 0;
  const int * m_linesize;

  const int m_align = 32;
};


